"""
"Copyright 2023 ClearBlade Inc."

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2023 ClearBlade Inc.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .utils import get_value
from .resources import HttpState, MqttState, LogLevel

class EventNotificationConfig:
    def __init__(self, pub_sub_topic_name, subfolder_matches=None) -> None:
        self._pub_sub_topic_name = pub_sub_topic_name
        self._sub_folder_matches = subfolder_matches

    @property
    def pub_sub_topic_name(self):
        return self._pub_sub_topic_name

    @property
    def sub_folder_matches(self):
        return self._sub_folder_matches

class DeviceRegistry:
    def __init__(self, id:str = None, name:str = None,
                 eventNotificationConfigs:list = [],
                 stateNotificationConfig:dict = {'pubsubTopicName': ''},
                 mqttConfig:dict = {'mqttEnabledState': MqttState.MQTT_ENABLED},
                 httpConfig:dict = {'httpEnabledState': HttpState.HTTP_ENABLED},
                 logLevel:str = LogLevel.NONE, credentials:list = []) -> None:
        self._id = id
        self._name = name
        self._event_notification_configs = eventNotificationConfigs
        self._state_notification_config = stateNotificationConfig
        self._mqtt_config = mqttConfig
        self._http_config = httpConfig
        self._loglevel = logLevel
        self._credentials = credentials

    @staticmethod
    def from_json(registry_json):
        event_notification_configs = []
        if "eventNotificationConfigs" in registry_json:
            event_notification_configs_json = registry_json['eventNotificationConfigs']

            for event_notification_config_json in event_notification_configs_json:
                if "subfolderMatches" in event_notification_config_json:
                    event_notification_config = EventNotificationConfig(event_notification_config_json['pubsubTopicName'], event_notification_config_json["subfolderMatches"])
                else:
                    event_notification_config = EventNotificationConfig(event_notification_config_json['pubsubTopicName'])
                event_notification_configs.append(event_notification_config)

        return DeviceRegistry(id=get_value(registry_json,'id'), name=get_value(registry_json,'name'),
                              eventNotificationConfigs=event_notification_configs,
                              stateNotificationConfig=get_value(registry_json,'stateNotificationConfig'),
                              mqttConfig=get_value(registry_json,'mqttConfig'),
                              httpConfig=get_value(registry_json,'httpConfig'),
                              credentials=get_value(registry_json,'credentials'),
                              logLevel=get_value(registry_json,'logLevel'))

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def event_notification_configs(self):
        return self._event_notification_configs

    @property
    def state_notification_config(self):
        return self._state_notification_config

    @property
    def mqtt_config(self):
        return self._mqtt_config

    @property
    def http_config(self):
        return self._http_config

    @property
    def log_level(self):
        return self._loglevel

    @property
    def credentials(self):
        return self._credentials
    
class CreateDeviceRegistryRequest:
    def __init__(self, parent:str = None,
                 device_registry:DeviceRegistry = None) -> None:
        self._parent = parent
        self._device_registry = device_registry

    @property
    def parent(self) -> str:
        return self._parent

    @property
    def device_registry(self)-> DeviceRegistry:
        return self._device_registry

class UpdateDeviceRegistryRequest:
    def __init__(self, name:str = None,
                 updateMask:str = None,
                 device_registry:DeviceRegistry = None) -> None:
        self._name = name
        self._update_mask = updateMask
        self._device_registry = device_registry

    @property
    def name(self) -> str:
        return self._name

    @property
    def update_mask(self) -> str:
        return self._update_mask

    @property
    def device_registry(self)-> DeviceRegistry:
        return self._device_registry

class GetDeviceRegistryRequest:
    def __init__(self, name:str = None) -> None:
        self._name = name
    
    @property
    def name(self):
        return self._name

class DeleteDeviceRegistryRequest:
    def __init__(self, name:str = None) -> None:
        self._name = name
    
    @property
    def name(self):
        return self._name


class ListDeviceRegistriesRequest:
    def __init__(self, parent:str = None,
                 page_size:int = None,
                 page_token:str = None) -> None:
        self._parent = parent
        self._page_size = page_size
        self._page_token = page_token

    @property
    def parent(self):
        return self._parent

    @property
    def page_size(self):
        return self._page_size

    @property
    def page_token(self):
        return self._page_token

class ListDeviceRegistriesResponse:
    def __init__(self, device_registries, next_page_token) -> None:
        self._device_registries = device_registries
        self._next_page_token = next_page_token

    @property
    def raw_page(self):
        return self

    @property
    def device_registries(self):
        return self._device_registries

    @property
    def next_page_token(self):
        return self._next_page_token

    @staticmethod
    def from_json(response_json):
        devices_registries_json = get_value(response_json, 'deviceRegistries')
        next_page_token = get_value(response_json, 'nextPageToken')
        device_registries = []

        for device_registry_json in devices_registries_json:
            device_registry = DeviceRegistry.from_json(registry_json=device_registry_json)
            device_registries.append(device_registry)

        return ListDeviceRegistriesResponse(device_registries, next_page_token)


"""
"Copyright 2023 ClearBlade Inc."

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2023 ClearBlade Inc.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import httpx
from .config import *

class HttpClient():
    def __init__(self,
                 clearblade_config:ClearBladeConfig = None) -> None:

        self._clearblade_config = clearblade_config
        self._system_key = self._clearblade_config.system_key
        self._auth_token = self._clearblade_config.token
        self._post_url : str = None
        self._request_headers : dict = None
        self._post_body : dict = None

    def _get_time_out(self):
        return httpx.Timeout(timeout=10.0, read=None)

    def _get_api_url(self, api_name:str = None, is_web_hook:bool = True) -> str:
        api_web_hook_path = "api/v/4/webhook/execute"
        if not is_web_hook:
            api_web_hook_path = "api/v/1/code"

        return "{}:{}/{}/{}/{}?".format(self._clearblade_config.api_url,
                                        "443",
                                        api_web_hook_path,
                                        self._system_key,
                                        api_name)

    def _headers(self):
        return {'ClearBlade-UserToken':self._clearblade_config.token, 'Content-Type':'application/json'}

    def _process_request_body(self, request_body = {}):
        return json.dumps(request_body)

    def get(self,api_name:str = None):
        self._post_url = self._get_api_url(api_name=api_name)
        self._request_headers = self._headers()

    def post(self, api_name:str = None,
             is_webhook_folder:bool = True,
             request_body:dict = {}):

        self._post_url = self._get_api_url(api_name=api_name, is_web_hook=is_webhook_folder)
        self._request_headers = self._headers()
        self._post_body = self._process_request_body(request_body=request_body)

    def delete(self,api_name:str = None):
        self._post_url = self._get_api_url(api_name=api_name)
        self._request_headers = self._headers()

    def patch(self, api_name:str = None, request_body:dict = {}):
        self._post_url = self._get_api_url(api_name=api_name)
        self._request_headers = self._headers()
        self._post_body = self._process_request_body(request_body=request_body)

class SyncClient(HttpClient):

    def __init__(self, clearblade_config: ClearBladeConfig = None) -> None:
        super().__init__(clearblade_config)
        self._sync_client = httpx.Client(timeout=self._get_time_out())

    def get(self, api_name:str = None, request_params:dict = {}):
        super().get(api_name=api_name)
        response = self._sync_client.get(url=self._post_url,
                                     headers=self._request_headers,
                                     params=request_params)
        self._sync_client.close()
        return response

    def post(self, api_name:str = None,
             is_webhook_folder:bool = True,
             request_params = {}, request_body = {}):
        super().post(api_name=api_name, is_webhook_folder=is_webhook_folder,
                     request_body=request_body)
        response = self._sync_client.post(url=self._post_url,
                                      headers=self._request_headers,
                                      params=request_params,
                                      data=self._post_body)
        self._sync_client.close()
        return response

    def delete(self, api_name:str = None, request_params:dict = None):
        super().delete(api_name = api_name)
        response = self._sync_client.delete(url=self._post_url,
                                        headers=self._request_headers,
                                        params=request_params)
        self._sync_client.close()
        return response

    def patch(self, api_name: str = None, request_body: dict = {}, request_params:dict = {}):
        super().patch(api_name, request_body)
        response = self._sync_client.patch(url=self._post_url,
                                           headers=self._request_headers,
                                           params = request_params,
                                           data=self._post_body)
        self._sync_client.close()
        return response

class AsyncClient(HttpClient):

    def __init__(self, clearblade_config: ClearBladeConfig = None) -> None:
        super().__init__(clearblade_config)
        self._async_client = httpx.AsyncClient(timeout=self._get_time_out())

    async def get(self, api_name:str = None, request_params:dict = {}):
        super().get(api_name=api_name)
        response = await self._async_client.get(url=self._post_url,
                                            headers=self._request_headers,
                                            params=request_params)
        await self._async_client.aclose()
        return response

    async def post(self, api_name:str = None,
                   is_webhook_folder:bool = True,
                   request_params = {}, request_body={}):
        super().post(api_name=api_name, is_webhook_folder=is_webhook_folder,
                     request_body=request_body)
        response = await self._async_client.post(url=self._post_url,
                                             headers=self._request_headers,
                                             params=request_params,
                                             data=self._post_body)
        await self._async_client.aclose()
        return response

    async def delete(self, api_name:str = None, request_params:dict = {}):
        super().delete(api_name=api_name)
        return await self._async_client.delete(url=self._post_url,
                                               headers=self._request_headers,
                                               params=request_params)

    async def patch(self, api_name:str = None, request_params:dict = {}, request_body:dict = {}):
        super().patch(api_name = api_name, request_body=request_body)
        return await self._async_client.patch(url=self._post_url,
                                              headers=self._request_headers,
                                              params = request_params,
                                              data=self._post_body)



"""
"Copyright 2023 ClearBlade Inc."

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2023 ClearBlade Inc.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from client import DeviceManagerClient, DeviceManagerAsyncClient
from device_types import *
from registry import *
from registry_types import *
import asyncio

def test_send_command():
    client = DeviceManagerClient()
    request = SendCommandToDeviceRequest(name='python_1', binary_data=b'QUJD')
    response = client.send_command_to_device(request)
    print(response)

async def test_send_command_async():
    async_client = DeviceManagerAsyncClient()
    request = SendCommandToDeviceRequest(name='python_1', binary_data=b'R2FyZ2l0ZXN0aW5n')
    response = await async_client.send_command_to_device(request=request)
    print(response)

def test_create_device():
    client = DeviceManagerClient()
    device = Device(id="Python_101", name="Python_101")
    request = CreateDeviceRequest(device=device)
    response = client.create_device(request=request)
    print(response)

async def test_create_device_async():
    async_client = DeviceManagerAsyncClient()
    device = Device(id="Python_12", name="Python_12")
    request = CreateDeviceRequest(device=device)
    response = await async_client.create_device(request=request)
    print(response)

def test_modify_cloud_to_device_config():
    client =  DeviceManagerClient()
    modify_cloud_config_device_request = ModifyCloudToDeviceConfigRequest(name='python_1', binary_data=b'QUJD', version_to_update=1)
    response = client.modify_cloud_to_device_config(request=modify_cloud_config_device_request)
    print(response)

async def test_modify_cloud_to_device_config_async():
    async_client =  DeviceManagerAsyncClient()
    modify_cloud_config_device_request = ModifyCloudToDeviceConfigRequest(name='Python_101', binary_data=b'QUJD', version_to_update=2)
    response = await async_client.modify_cloud_to_device_config(request=modify_cloud_config_device_request)
    print(response)

def test_delete_device():
    client =  DeviceManagerClient()
    delete_device_request = DeleteDeviceRequest(name='Python_12')
    response = client.delete_device(request=delete_device_request)
    print(response)

async def test_delete_device_async():
    async_client =  DeviceManagerAsyncClient()
    delete_device_request = DeleteDeviceRequest(name='Python_10')
    response = await async_client.delete_device(request=delete_device_request)
    print(response)

def test_get_device():
    client =  DeviceManagerClient()
    get_device_request = GetDeviceRequest(name='Python_101')
    response = client.get_device(request=get_device_request)
    print(response)

async def test_get_device_async():
    async_client =  DeviceManagerAsyncClient()
    get_device_request = GetDeviceRequest(name='Python_101')
    response = await async_client.get_device(request=get_device_request)
    print(response)

def test_bind_gateway_device():
    client =  DeviceManagerClient()
    bind_device_request = BindDeviceToGatewayRequest(deviceId='Python_101',gatewayId='gateway1')
    response = client.bind_device_to_gateway(request=bind_device_request)
    print(response)

async def test_bind_gateway_device_async():
    async_client =  DeviceManagerAsyncClient()
    bind_device_request = BindDeviceToGatewayRequest(deviceId='mandar_device',gatewayId='gateway1')
    response = await async_client.bind_device_to_gateway(request=bind_device_request)
    print(response)

def test_unbind_gateway_device():
    client =  DeviceManagerClient()
    bind_device_request = UnbindDeviceFromGatewayRequest(deviceId='Python_101',gatewayId='gateway1')
    response = client.unbind_device_from_gateway(request=bind_device_request)
    print(response)

async def test_unbind_gateway_device_async():
    async_client =  DeviceManagerAsyncClient()
    bind_device_request = UnbindDeviceFromGatewayRequest(deviceId='mandar_device',gatewayId='gateway1')
    response = await async_client.unbind_device_from_gateway(request=bind_device_request)
    print(response)

def test_get_device_states():
    client = DeviceManagerClient()
    request = ListDeviceStatesRequest(name='Python_5', num_states=3)
    response = client.list_device_states(request)
    print(response)

async def test_get_device_states_async():
    async_client = DeviceManagerAsyncClient()
    request = ListDeviceStatesRequest(name='Python_6', num_states=3)
    response = await async_client.list_device_states(request=request)
    print(response)

def test_get_device_configVersions():
    client = DeviceManagerClient()
    request = ListDeviceConfigVersionsRequest(name='Python_6', numVersions=1)
    response = client.list_device_config_versions(request)
    device_configs = response.device_configs
    for device_config in device_configs:
        print("Device version = {} Device Ack Time {} \n".format(device_config.version,
        device_config.cloud_update_time))

async def test_get_device_configVersions_async():
    async_client = DeviceManagerAsyncClient()
    request = ListDeviceConfigVersionsRequest(name='Python_6', numVersions=1)
    response = await async_client.list_device_config_versions(request=request)
    print(response)

def test_get_devices_list():
    client =  DeviceManagerClient()
    get_devices_list_request = ListDevicesRequest(parent='projects/ingressdevelopmentenv/locations/us-central1', pageSize=2)
    page_result = client.list_devices(request=get_devices_list_request)
    for response in page_result:
       print(response)

async def test_get_devices_list_async():
    async_client = DeviceManagerAsyncClient()
    get_devices_list_request = ListDevicesRequest(parent='projects/ingressdevelopmentenv/locations/us-central1')
    page_result = await async_client.list_devices(request=get_devices_list_request)
    async for response in page_result:
        print(response)

def test_update_device():
    client =  DeviceManagerClient()
    update_device_request = UpdateDeviceRequest(name='Rashmi_Device_Test',id='Rashmi_Device_Test',logLevel='NONE',blocked=True, updateMask='logLevel')
    response = client.update_device(request=update_device_request)
    print(response)

async def test_update_device_async():
    async_client = DeviceManagerAsyncClient()
    update_device_request = UpdateDeviceRequest(name='Rashmi_Device_Test',id='Rashmi_Device_Test',logLevel='ERROR',blocked=True, updateMask='logLevel')
    response = await async_client.update_device(request=update_device_request)
    print(response)

def test_list_registries():
    client = DeviceManagerClient()
    request = ListDeviceRegistriesRequest(parent="projects/ingressdevelopmentenv/locations/us-central1")
    response = client.list_device_registries(request=request)
    for page_result in response:
        print(page_result)

async def test_list_registries_async():
    client = DeviceManagerAsyncClient()
    request = ListDeviceRegistriesRequest(parent="projects/ingressdevelopmentenv/locations/us-central1")
    response = await client.list_device_registries(request=request)
    async for page_result in response:
        print(page_result)

def test_get_registry():
    client = DeviceManagerClient()
    request = GetDeviceRegistryRequest()
    response = client.get_device_registry(request=request)
    print(response)

async def test_get_registry_async():
    async_client = DeviceManagerAsyncClient()
    request = GetDeviceRegistryRequest()
    response = await async_client.get_device_registry(request=request)
    print(response)

def test_create_registry():
    client = DeviceManagerClient()
    registry = DeviceRegistry(id='deleteTest5', mqttConfig={'mqttEnabledState':'MQTT_ENABLED'},
                              httpConfig={'httpEnabledState':'HTTP_ENABLED'},
                              logLevel='ERROR', 
                              eventNotificationConfigs=[{'pubsubTopicName':'projects/ingressdevelopmentenv/topics/deleting'}])
    request = CreateDeviceRegistryRequest(parent="projects/ingressdevelopmentenv/locations/us-central1",device_registry=registry)
    response = client.create_device_registry(request=request)
    print(response)

async def test_create_registry_async():
    async_client = DeviceManagerAsyncClient()
    registry = DeviceRegistry(id='deletetest3', name='deletetest3', mqttConfig={'mqttEnabledState':'MQTT_ENABLED'},httpConfig={'httpEnabledState':'HTTP_ENABLED'},logLevel='ERROR', eventNotificationConfigs=[{'pubsubTopicName':'projects/ingressdevelopmentenv/topics/deleting'}])
    request = CreateDeviceRegistryRequest(parent="projects/ingressdevelopmentenv/locations/us-central1",device_registry=registry)
    response = await async_client.create_device_registry(request=request)
    print(response)

def test_delete_registry():
    client = DeviceManagerClient()
    request = DeleteDeviceRegistryRequest(name="projects/ingressdevelopmentenv/locations/us-central1/registries/deleteTest2")
    response = client.delete_device_registry(request=request)
    print(response)

async def test_delete_registry_async():
    async_client = DeviceManagerAsyncClient()
    request = DeleteDeviceRegistryRequest(name="projects/ingressdevelopmentenv/locations/us-central1/registries/deleteTest3")
    response = await async_client.delete_device_registry(request=request)
    print(response)

def test_update_registry():
    client = DeviceManagerClient()
    registry = DeviceRegistry(id='deleteTest5', name='deleteTest5', mqttConfig={'mqttEnabledState':'MQTT_DISABLED'},logLevel='ERROR')
    request = UpdateDeviceRegistryRequest(name="projects/ingressdevelopmentenv/locations/us-central1/registries/deleteTest5", updateMask='mqttConfig.mqtt_enabled_state,logLevel', device_registry=registry)
    response = client.update_device_registry(request=request)
    print(response)

async def test_update_registry_async():
    async_client = DeviceManagerAsyncClient()
    registry = DeviceRegistry(id='deleteTest5', name='deleteTest5', mqttConfig={'mqttEnabledState':'MQTT_ENABLED'},logLevel='DEBUG')
    request = UpdateDeviceRegistryRequest(name="projects/ingressdevelopmentenv/locations/us-central1/registries/deleteTest5", updateMask='mqttConfig.mqtt_enabled_state,logLevel', device_registry=registry)
    response = await async_client.update_device_registry(request=request)
    print(response)

if __name__ ==  '__main__':
    #test_send_command()
    #asyncio.run(test_send_command_async())
    #test_create_device()
    #asyncio.run(test_create_device_async())
    #test_modify_cloud_to_device_config()
    #asyncio.run(test_modify_cloud_to_device_config_async())
    #test_delete_device()
    #asyncio.run(test_delete_device_async())
    #test_get_device()
    #asyncio.run(test_get_device_async())
    #test_bind_gateway_device()
    #asyncio.run(test_bind_gateway_device_async())
    #test_unbind_gateway_device()
    #asyncio.run(test_unbind_gateway_device_async())
    #test_get_device_states()
    #asyncio.run(test_get_device_states_async())
    #test_get_device_configVersions()
    #asyncio.run(test_get_device_configVersions_async())
    
    test_list_registries()
    asyncio.run(test_list_registries_async())

    test_get_devices_list()
    asyncio.run(test_get_devices_list_async())
    
    #test_update_device()
    #asyncio.run(test_update_device_async())
    #test_list_registries()
    #test_get_registry()
    #asyncio.run(test_get_registry_async())
    #test_create_registry()
    #asyncio.run(test_create_registry_async())
    #test_delete_registry()
    #asyncio.run(test_delete_registry_async())
    #test_update_registry()
    #asyncio.run(test_update_registry_async())



"""
"Copyright 2023 ClearBlade Inc."

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2023 ClearBlade Inc.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

class ClearBladeConfig:
    def __init__(self, system_key:str = None,
                 auth_token:str = None,
                 api_url:str = None,
                 region:str = None,
                 project:str = None ) -> None:
        self._system_key = system_key
        self._auth_token = auth_token
        self._api_url = api_url
        self._region = region
        self._project = project

    @property
    def system_key(self):
        return self._system_key

    @property
    def token(self):
        return self._auth_token

    @property
    def api_url(self):
        return self._api_url

    @property
    def project(self):
        return self._project

    @property
    def region(self):
        return self._region



"""
"Copyright 2023 ClearBlade Inc."

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2023 ClearBlade Inc.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import os

from .config import ClearBladeConfig
from .http_client import AsyncClient, SyncClient
from .utils import find_project_region_registry_from_parent


class ClearBladeConfigManager:
    def __init__(self) -> None:
        self._admin_config: ClearBladeConfig = None
        self._regional_config: ClearBladeConfig = None
        self._registry_name: str = None
        self._region_name: str = None

    def _set_admin_clearblade_config(self):
        if self._admin_config:
            return

        service_account_file_path = os.environ.get("CLEARBLADE_CONFIGURATION")
        service_account_data = None
        
        if service_account_file_path is None:
            raise Exception('CLEARBLADE_CONFIGURATION environment variable is not set')
        
        #parse the file and get all te required details.
        with open(service_account_file_path, mode='r') as service_account_file:
            service_account_data = json.load(service_account_file)

        if service_account_data is None:
            raise Exception('ClearBlade Service account file is empty')

        system_key = service_account_data['systemKey']
        auth_token = service_account_data['token']
        api_url = service_account_data['url']
        project = service_account_data['project']

        self._admin_config = ClearBladeConfig(system_key=system_key, auth_token=auth_token,
                                                 api_url=api_url, project=project)

    def _create_regional_config_from_env(self, systemKey:str = None, serviceAccountToken:str = None, url:str = None, region:str = None)-> ClearBladeConfig :
        return ClearBladeConfig(system_key=systemKey, auth_token=serviceAccountToken, api_url=url,
                                region=region, project=self._admin_config.project)

    
    def _create_regional_config(self, regional_json: json = None)-> ClearBladeConfig :
        system_key = regional_json['systemKey']
        auth_token = regional_json['serviceAccountToken']
        api_url = regional_json['url']
        region = regional_json['region']

        return ClearBladeConfig(system_key=system_key, auth_token=auth_token, api_url=api_url,
                                region=region, project=self._admin_config.project)

    def _set_regional_config(self, region:str = None, registry:str = None):
        self._set_admin_clearblade_config()

        if not region:
            region = self.region_name

        if not registry:
            registry = self.registry_name

        if not region or not registry:
            raise Exception("Either location or registry name is not provided")
        
        system_key = os.environ.get("REGISTRY_SYSKEY")
        registry_token = os.environ.get("REGISTRY_TOKEN")
        url = os.environ.get("REGISTRY_URL")
        if (system_key and registry_token and url) :
            self._regional_config = self._create_regional_config_from_env(systemKey=system_key , serviceAccountToken = registry_token, url = url, region = region)
        else :
            request_body = {'region':region,'registry':registry, 'project':self._admin_config.project}
            sync_client = SyncClient(clearblade_config=self._admin_config)
            response = sync_client.post(api_name="getRegistryCredentials", is_webhook_folder=False,
                                        request_body=request_body)
            if response.status_code != 200:
                raise Exception(
                    f"\n\nRegistry Information not found! Please check if the given registry exists\nProject: {self._admin_config.project}\nRegistry: {registry}\nRegion: {region}"
                    )
            response_json = response.json()
            response_json['region'] = region
            self._regional_config = self._create_regional_config(regional_json=response_json) 
            

    async def _set_regional_config_async(self, region:str = None, registry:str = None):

        self._set_admin_clearblade_config()

        if not region:
            region = self._region_name

        if not registry:
            registry = self.registry_name
        system_key = os.environ.get("REGISTRY_SYSKEY")
        registry_token = os.environ.get("REGISTRY_TOKEN")
        url = os.environ.get("REGISTRY_URL")
        if (system_key and registry_token) :
            self._regional_config = self._create_regional_config_from_env(systemKey=system_key , serviceAccountToken = registry_token, url = url, region = region)
        else :            
            async_client = AsyncClient(clearblade_config=self._admin_config)
            request_body = {'region':region,'registry':registry, 'project':self._admin_config.project}
            response = await async_client.post(api_name="getRegistryCredentials",
                                            is_webhook_folder=False,
                                            request_body=request_body)

            if response.status_code != 200:
                raise Exception(
                    f"\n\nRegistry Information not found! Please check if the given registry exists\nProject: {self._admin_config.project}\nRegistry: {registry}\nRegion: {region}"
                    )
            response_json = response.json()
            response_json['region'] = region
            self._regional_config = self._create_regional_config(regional_json=response_json)


    @property
    def admin_config(self):
        if not self._admin_config:
            self._set_admin_clearblade_config()

        return self._admin_config

    def regional_config(self, parent: str):
        if not self._regional_config:
            project = find_project_region_registry_from_parent(parent)
            self._set_regional_config(region=project['location'], registry=project['registry'])

        return self._regional_config

    async def regional_config_async(self, parent: str):
        if not self._regional_config:
            project = find_project_region_registry_from_parent(parent)
            await self._set_regional_config_async(region=project['location'], registry=project['registry'])

        return self._regional_config

    @property
    def registry_name(self):
        return self._registry_name

    @property
    def region_name(self):
        return self._region_name

    @registry_name.setter
    def registry_name(self, registry_name):
        self._registry_name = registry_name

    @region_name.setter
    def region_name(self, region_name):
        self._region_name = region_name



"""
"Copyright 2023 ClearBlade Inc."

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2023 ClearBlade Inc.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .config_manager import ClearBladeConfigManager
from .device_types import *
from .http_client import AsyncClient, SyncClient
from .pagers import ListDevicesAsyncPager, ListDevicesPager
import base64
from .resources import DeviceCredential

class ClearBladeDeviceManager():

    def __init__(self) -> None:
        # create the ClearBladeConfig object
        self._config_manager = ClearBladeConfigManager()

    def _prepare_for_send_command(self,
                                  request: SendCommandToDeviceRequest,
                                  name: str = None,
                                  binary_data: bytes = None,
                                  subfolder: str = None):
        has_flattened_params = any([name, binary_data])
        if request is not None and has_flattened_params:
            raise ValueError(
                "If the `request` argument is set, then none of "
                "the individual field arguments should be set."
            )

        if request is None:
            request = SendCommandToDeviceRequest(name, binary_data, subfolder)
        params = {'name':request.parent,'method':'sendCommandToDevice'}
        body = {'binaryData':base64.b64encode(request.binary_data).decode("utf-8"), 'subfolder':request.sub_folder}

        return params,body

    def _create_device_body(self, device: Device) :
        return {'id':device.id,
                'credentials':DeviceCredential.convert_credentials_for_create_update(device.credentials),
                'config':device.config,
                'blocked': device.blocked,
                'logLevel':device.log_level, 'metadata':device.meta_data,
                'gatewayConfig':device.gateway_config}

    def _create_device_from_response(self, json_response) -> Device :
        return Device.from_json(json=json_response)

    def _prepare_modify_cloud_config_device(self,
                                            request: ModifyCloudToDeviceConfigRequest,
                                            name, binary_data, version_to_update):

        has_flattened_params = any([name, binary_data])
        if request is not None and has_flattened_params:
            raise ValueError(
                "If the `request` argument is set, then none of "
                "the individual field arguments should be set."
            )

        if request is None:
            request = ModifyCloudToDeviceConfigRequest(name=name,version_to_update=version_to_update, binary_data=binary_data)

        params = {'name': request.parent, 'method': 'modifyCloudToDeviceConfig'}
        body = {'binaryData': request.binary_data.decode("utf-8"), 'versionToUpdate':request.version_to_update}

        return params, body

    def _create_device_config(self, response):
        return DeviceConfig.from_json(response)

    def _create_device_list_from_response(self, json_response):
        devicesList = []
        for deviceJson in json_response['devices']:
            devicesList.append(Device.from_json(json=deviceJson))
        return devicesList

    def _set_project_location_region(self, data):
        if 'locations' in data:
            self._config_manager.region_name = data['locations']

        if 'registries' in data:
            self._config_manager.registry_name = data['registries']

    def device_path(self,
                    project,
                    location,
                    registry,
                    device) -> str:
        """Returns a fully-qualified device string."""

        return "projects/{project}/locations/{location}/registries/{registry}/devices/{device}".format(
            project=project,
            location=location,
            registry=registry,
            device=device,
        )

    def registry_path(self,
                      project: str,
                      location: str,
                      registry: str,
                      ) -> str:
        """Returns a fully-qualified registry string."""

        return "projects/{project}/locations/{location}/registries/{registry}".format(
            project=project,
            location=location,
            registry=registry,
        )

    def send_command(self,
                     request: SendCommandToDeviceRequest,
                     name: str = None,
                     binary_data: bytes = None,
                     subfolder: str = None):
        if not request.parent:
            raise Exception('Device path must be supplied to the request object.')

        clearblade_config = self._config_manager.regional_config(parent=request.parent)

        params, body = self._prepare_for_send_command(request, name, binary_data, subfolder)
        sync_client = SyncClient(clearblade_config=clearblade_config)
        return sync_client.post(api_name="cloudiot_devices", request_params=params, request_body=body)

    async def send_command_async(self,
                                 request: SendCommandToDeviceRequest = None,
                                 name: str = None,
                                 binary_data: str = None,
                                 subfolder: str = None):

        if not request.parent:
            raise Exception('Device path must be supplied to the request object.')

        params, body = self._prepare_for_send_command(request, name, binary_data, subfolder)
        clearblade_config = await self._config_manager.regional_config_async(request.parent)
        async_client = AsyncClient(clearblade_config=clearblade_config)
        return await async_client.post(api_name="cloudiot_devices",
                                       request_params=params,
                                       request_body=body)

    def create(self, request: CreateDeviceRequest) -> Device:
        body = self._create_device_body(request.device)

        if not request.parent:
            raise Exception('Parent must be supplied to the request object.')

        clearblade_config = self._config_manager.regional_config(parent=request.parent)

        sync_client = SyncClient(clearblade_config=clearblade_config)
        response = sync_client.post(api_name="cloudiot_devices",
                                    request_body=body)

        # return the device only if status code is 200
        # other wise return None
        if response.status_code == 200:
            return self._create_device_from_response(response.json())
        return response.json()

    async def create_async(self, request: CreateDeviceRequest,
                           parent: str = None,
                           device: Device = None) -> Device:
        if not request.parent:
            raise Exception('Parent must be supplied to the request object.')

        body = self._create_device_body(request.device)
        clearblade_config = await self._config_manager.regional_config_async(request.parent)
        async_client = AsyncClient(clearblade_config=clearblade_config)
        response = await async_client.post(api_name="cloudiot_devices", request_body=body)

        if response.status_code == 200:
            return self._create_device_from_response(response.json())
        return response.json()

    def modify_cloud_device_config(self,
                                   request: ModifyCloudToDeviceConfigRequest,
                                   name:str = None,
                                   version_to_update : int = None,
                                   binary_data: bytes = None):
        if not request.parent:
            raise Exception('Device path must be supplied to the request object.')

        clearblade_config = self._config_manager.regional_config(parent=request.parent)

        params, body = self._prepare_modify_cloud_config_device(request=request, name=name,
                                                                binary_data=binary_data,
                                                                version_to_update=version_to_update)
        sync_client = SyncClient(clearblade_config=clearblade_config)
        response = sync_client.post(api_name="cloudiot_devices", request_params=params, request_body=body)

        if response.status_code == 200:
            return self._create_device_config(response.json())

        return response.json()

    async def modify_cloud_device_config_async(self,
                                               request: ModifyCloudToDeviceConfigRequest,
                                               name:str = None,
                                               version_to_update : int = None,
                                               binary_data: bytes = None):
        if not request.parent:
            raise Exception('Device path must be supplied to the request object.')

        params, body = self._prepare_modify_cloud_config_device(request=request, name=name,
                                                                binary_data=binary_data,
                                                                version_to_update=version_to_update)

        clearblade_config = await self._config_manager.regional_config_async(request.parent)
        async_client = AsyncClient(clearblade_config=clearblade_config)
        response = await async_client.post(api_name="cloudiot_devices",
                                           request_params=params,
                                           request_body=body)
        if response.status_code == 200:
            return self._create_device_config(response.json())
        return response

    def get_device_list_response(self, request:ListDevicesRequest):
        if not request.parent:
            raise Exception('Registry path must be supplied to the request object.')

        clearblade_config = self._config_manager.regional_config(parent=request.parent)

        sync_client = SyncClient(clearblade_config=clearblade_config)

        params = request._prepare_params_for_list()
        response = sync_client.get(api_name="cloudiot_devices",request_params=params)

        if response.status_code == 200:
            return ListDevicesResponse.from_json(response.json())
        return response

    def list(self,
             request: ListDevicesRequest):

        device_list_response = self.get_device_list_response(request=request)
        if device_list_response:
            return ListDevicesPager(method=self.get_device_list_response,
                                    request=request, response=device_list_response)
        return None

    async def get_device_list_async(self, request: ListDevicesRequest):

        if not request.parent:
            raise Exception('Registry path must be supplied to the request object.')

        params = request._prepare_params_for_list()
        clearblade_config = await self._config_manager.regional_config_async(request.parent)
        async_client = AsyncClient(clearblade_config=clearblade_config)
        response = await async_client.get(api_name="cloudiot_devices",request_params=params)

        if response.status_code == 200:
            return ListDevicesResponse.from_json(response.json())
        return response

    async def list_async(self,
                         request: ListDevicesRequest):
        device_list_response = await self.get_device_list_async(request=request)
        if device_list_response:
            return ListDevicesAsyncPager(method=self.get_device_list_async,
                                         request=request, response=device_list_response)
        return None

    def get(self,
            request: GetDeviceRequest) -> Device:

        if not request.parent:
            raise Exception('Device path must be supplied to the request object.')

        clearblade_config = self._config_manager.regional_config(parent=request.parent)

        params = {'name':request.parent}
        sync_client = SyncClient(clearblade_config=clearblade_config)
        response = sync_client.get(api_name="cloudiot_devices", request_params=params)

        if response.status_code == 200:
            return self._create_device_from_response(response.json())
        return response

    async def get_async(self,
                        request: GetDeviceRequest) -> Device:
        if not request.parent:
            raise Exception('Device path must be supplied to the request object.')

        params = {'name':request.parent}
        clearblade_config = await self._config_manager.regional_config_async(request.parent)
        async_client = AsyncClient(clearblade_config=clearblade_config)
        response = await async_client.get(api_name="cloudiot_devices", request_params=params)

        if response.status_code == 200:
            return self._create_device_from_response(response.json())

        return response

    def delete(self,
               request: DeleteDeviceRequest):
        if not request.parent:
            raise Exception('Parent must be supplied to the request object.')

        clearblade_config = self._config_manager.regional_config(parent=request.parent)

        params = {'name':request.parent}

        sync_client = SyncClient(clearblade_config=clearblade_config)
        response = sync_client.delete(api_name="cloudiot_devices",request_params=params)
        return response

    async def delete_async(self,
                           request: DeleteDeviceRequest):

        if not request.parent:
            raise Exception('Device path must be supplied to the request object.')

        params = {'name':request.parent}
        clearblade_config = await self._config_manager.regional_config_async(request.parent)
        async_client = AsyncClient(clearblade_config=clearblade_config)
        response = await async_client.delete(api_name="cloudiot_devices", request_params=params)
        return response

    def update(self,
               request: UpdateDeviceRequest) -> Device:
        if not request.parent:
            raise Exception('Device path must be supplied to the request object.')

        clearblade_config = self._config_manager.regional_config(parent=request.parent)

        params, body = request._prepare_params_body_for_update()
        sync_client = SyncClient(clearblade_config=clearblade_config)
        response = sync_client.patch(api_name="cloudiot_devices",request_body=body, request_params=params)

        if response.status_code == 200:
            return self._create_device_from_response(json_response=response.json())
        return response

    async def update_async(self,
                           request: UpdateDeviceRequest) -> Device:
        if not request.parent:
            raise Exception('Parent must be supplied to the request object.')

        params, body = request._prepare_params_body_for_update()
        clearblade_config = await self._config_manager.regional_config_async(request.parent)
        async_client = AsyncClient(clearblade_config=clearblade_config)
        response = await async_client.patch(api_name="cloudiot_devices",request_body=body, request_params=params)

        if response.status_code == 200:
            return self._create_device_from_response(json_response=response.json())
        return response

    def bindGatewayToDevice(self,
                            request: BindDeviceToGatewayRequest) :
        body = {'deviceId':request.deviceId, 'gatewayId':request.gatewayId}
        params = {'method':'bindDeviceToGateway'}

        if not request.parent:
            raise Exception('Parent must be supplied to the request object.')

        clearblade_config = self._config_manager.regional_config(parent=request.parent)
        sync_client = SyncClient(clearblade_config=clearblade_config)
        response = sync_client.post(api_name="cloudiot", request_params=params, request_body=body)

        return response

    async def bindGatewayToDevice_async(self,
                                        request: BindDeviceToGatewayRequest) :
        if not request.parent:
            raise Exception('Parent must be supplied to the request object.')

        body = {'deviceId':request.deviceId, 'gatewayId':request.gatewayId}
        params = {'method':'bindDeviceToGateway'}
        clearblade_config = await self._config_manager.regional_config_async(request.parent)
        async_client = AsyncClient(clearblade_config=clearblade_config)
        response = await async_client.post(api_name="cloudiot", request_params=params, request_body=body)
        return response

    def unbindGatewayFromDevice(self,
                                request: UnbindDeviceFromGatewayRequest) :
        if not request.parent:
            raise Exception('Parent must be supplied to the request object.')

        clearblade_config = self._config_manager.regional_config(request.parent)
        sync_client = SyncClient(clearblade_config=clearblade_config)
        body = {'deviceId':request.deviceId, 'gatewayId':request.gatewayId}
        params = {'method':'unbindDeviceFromGateway'}
        response = sync_client.post(api_name="cloudiot", request_params=params, request_body=body)
        return response

    async def unbindGatewayFromDevice_async(self,
                                            request: UnbindDeviceFromGatewayRequest) :
        if not request.parent:
            raise Exception('Parent must be supplied to the request object.')

        body = {'deviceId':request.deviceId, 'gatewayId':request.gatewayId}
        params = {'method':'unbindDeviceFromGateway'}

        clearblade_config = await self._config_manager.regional_config_async(request.parent)
        async_client = AsyncClient(clearblade_config=clearblade_config)
        response = await async_client.post(api_name="cloudiot",request_params=params, request_body=body)
        return response

    def states_list(self,
                    request: ListDeviceStatesRequest):
        if not request.parent:
            raise Exception('Device Path must be supplied to the request object.')

        clearblade_config = self._config_manager.regional_config(request.parent)

        params = {'name':request.parent, 'numStates':request.numStates}
        sync_client = SyncClient(clearblade_config=clearblade_config)
        response = sync_client.get(api_name="cloudiot_devices_states",request_params=params)

        if response.status_code == 200:
            return ListDeviceStatesResponse.from_json(response.json())
        return response

    async def states_list_async(self,
                                request: ListDeviceStatesRequest):

        if not request.parent:
            raise Exception('Device Path must be supplied to the request object.')

        params = {'name':request.parent, 'numStates':request.numStates}
        clearblade_config = await self._config_manager.regional_config_async(request.parent)
        async_client = AsyncClient(clearblade_config=clearblade_config)
        response = await async_client.get(api_name="cloudiot_devices_states", request_params=params)

        if response.status_code == 200:
            return ListDeviceStatesResponse.from_json(response.json())
        return response

    def config_versions_list(self,
                             request: ListDeviceConfigVersionsRequest):
        if not request.parent:
            raise Exception('Device Path must be supplied to the request object.')

        clearblade_config = self._config_manager.regional_config(request.parent)
        params = {'name':request.parent, 'numVersions':request.numVersions}
        sync_client = SyncClient(clearblade_config=clearblade_config)
        response = sync_client.get(api_name="cloudiot_devices_configVersions",request_params=params)

        if response.status_code == 200:
            return ListDeviceConfigVersionsResponse.from_json(response.json())
        return response

    async def config_versions_list_async(self,
                                         request: ListDeviceConfigVersionsRequest):
        if not request.parent:
            raise Exception('Device Path must be supplied to the request object.')

        params = {'name':request.parent, 'numVersions':request.numVersions}
        clearblade_config = await self._config_manager.regional_config_async(request.parent)
        async_client = AsyncClient(clearblade_config=clearblade_config)
        response = await async_client.get(api_name="cloudiot_devices_configVersions", request_params=params)

        if response.status_code == 200:
            return ListDeviceConfigVersionsResponse.from_json(response.json())
        return response


"""
"Copyright 2023 ClearBlade Inc."

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2023 ClearBlade Inc.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .device_types import *
from .devices import *
from .registry import *
from .registry_types import *

class Client():
    def device_path(
        self,
        project: str,
        location: str,
        registry: str,
        device: str
        ) -> str:
        cb_device_manager = ClearBladeDeviceManager()
        return cb_device_manager.device_path(project=project, 
                                             location=location, 
                                             registry=registry, 
                                             device=device)
    
    def registry_path(
        self,
        project: str,
        location: str,
        registry: str
        ) -> str:
        cb_device_manager = ClearBladeDeviceManager()
        return cb_device_manager.registry_path(project=project, 
                                               location=location, 
                                               registry=registry)

class DeviceManagerClient(Client):

    def send_command_to_device(self,request:SendCommandToDeviceRequest):
        cb_device_manager = ClearBladeDeviceManager()
        return cb_device_manager.send_command(request)

    def create_device(self, request):
        cb_device_manager = ClearBladeDeviceManager()
        return cb_device_manager.create(request=request)

    def modify_cloud_to_device_config(self, request: ModifyCloudToDeviceConfigRequest,
                                      name: str = None,
                                      version_to_update = "",
                                      binary_data:bytes = None):

        cb_device_manager = ClearBladeDeviceManager()
        return cb_device_manager.modify_cloud_device_config(request=request,
                                                            name=name,
                                                            version_to_update=version_to_update,
                                                            binary_data=binary_data)

    def delete_device(self, request):
        cb_device_manager = ClearBladeDeviceManager()
        return cb_device_manager.delete(request=request)

    def get_device(self, request):
        cb_device_manager = ClearBladeDeviceManager()
        return cb_device_manager.get(request=request)

    def bind_device_to_gateway(self, request : BindDeviceToGatewayRequest):
        cb_device_manager = ClearBladeDeviceManager()
        return cb_device_manager.bindGatewayToDevice(request=request)

    def unbind_device_from_gateway(self, request : UnbindDeviceFromGatewayRequest):
        cb_device_manager = ClearBladeDeviceManager()
        return cb_device_manager.unbindGatewayFromDevice(request=request)

    def list_device_states(self, request : ListDeviceStatesRequest):
        cb_device_manager = ClearBladeDeviceManager()
        return cb_device_manager.states_list(request=request)

    def list_device_config_versions(self, request : ListDeviceConfigVersionsRequest):
        cb_device_manager = ClearBladeDeviceManager()
        return cb_device_manager.config_versions_list(request=request)

    def list_devices(self, request : ListDevicesRequest):
        cb_device_manager = ClearBladeDeviceManager()
        return cb_device_manager.list(request=request)

    def update_device(self, request : UpdateDeviceRequest):
        cb_device_manager = ClearBladeDeviceManager()
        return cb_device_manager.update(request=request)

    def list_device_registries(self, request: ListDeviceRegistriesRequest):
        cb_registry_manager = ClearBladeRegistryManager()
        return cb_registry_manager.list(request=request)

    def get_device_registry(self, request=GetDeviceRegistryRequest):
        cb_registry_manager = ClearBladeRegistryManager()
        return cb_registry_manager.get(request=request)

    def create_device_registry(self, request=CreateDeviceRegistryRequest):
        cb_registry_manager = ClearBladeRegistryManager()
        return cb_registry_manager.create(request=request)

    def delete_device_registry(self, request=DeleteDeviceRegistryRequest):
        cb_registry_manager = ClearBladeRegistryManager()
        return cb_registry_manager.delete(request=request)

    def update_device_registry(self, request=UpdateDeviceRegistryRequest):
        cb_registry_manager = ClearBladeRegistryManager()
        return cb_registry_manager.patch(request=request)

class DeviceManagerAsyncClient(Client):

    async def send_command_to_device(self, request:SendCommandToDeviceRequest):
        cb_device_manager = ClearBladeDeviceManager()
        return await cb_device_manager.send_command_async(request)

    async def create_device(self, request):
        cb_device_manager = ClearBladeDeviceManager()
        return await cb_device_manager.create_async(request)

    async def modify_cloud_to_device_config(self, request: ModifyCloudToDeviceConfigRequest,
                                            name: str = None,
                                            version_to_update = "",
                                            binary_data:bytes = None):

        cb_device_manager = ClearBladeDeviceManager()
        return await cb_device_manager.modify_cloud_device_config_async(request=request,
                                                            name=name,
                                                            version_to_update=version_to_update,
                                                            binary_data=binary_data)

    async def delete_device(self, request):
        cb_device_manager = ClearBladeDeviceManager()
        return await cb_device_manager.delete_async(request=request)

    async def get_device(self, request):
        cb_device_manager = ClearBladeDeviceManager()
        return await cb_device_manager.get_async(request=request)

    async def bind_device_to_gateway(self, request : BindDeviceToGatewayRequest):
        cb_device_manager = ClearBladeDeviceManager()
        return await cb_device_manager.bindGatewayToDevice_async(request=request)

    async def unbind_device_from_gateway(self, request : UnbindDeviceFromGatewayRequest):
        cb_device_manager = ClearBladeDeviceManager()
        return await cb_device_manager.unbindGatewayFromDevice_async(request=request)

    async def list_device_states(self, request : ListDeviceStatesRequest):
        cb_device_manager = ClearBladeDeviceManager()
        return await cb_device_manager.states_list_async(request=request)

    async def list_device_config_versions(self, request : ListDeviceConfigVersionsRequest):
        cb_device_manager = ClearBladeDeviceManager()
        return await cb_device_manager.config_versions_list_async(request=request)

    async def list_devices(self, request : ListDevicesRequest):
        cb_device_manager = ClearBladeDeviceManager()
        return await cb_device_manager.list_async(request=request)

    async def update_device(self, request : UpdateDeviceRequest):
        cb_device_manager = ClearBladeDeviceManager()
        return await cb_device_manager.update_async(request=request)

    async def list_device_registries(self, request: ListDeviceRegistriesRequest):
        cb_registry_manager = ClearBladeRegistryManager()
        return await cb_registry_manager.list_async(request=request)

    async def get_device_registry(self, request=GetDeviceRegistryRequest):
        cb_registry_manager = ClearBladeRegistryManager()
        return await cb_registry_manager.get_async(request=request)

    async def create_device_registry(self, request=CreateDeviceRegistryRequest):
        cb_registry_manager = ClearBladeRegistryManager()
        return await cb_registry_manager.create_async(request=request)

    async def delete_device_registry(self, request=DeleteDeviceRegistryRequest):
        cb_registry_manager = ClearBladeRegistryManager()
        return await cb_registry_manager.delete_async(request=request)

    async def update_device_registry(self, request=UpdateDeviceRegistryRequest):
        cb_registry_manager = ClearBladeRegistryManager()
        return await cb_registry_manager.patch_async(request=request)



"""
"Copyright 2023 ClearBlade Inc."

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2023 ClearBlade Inc.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List
from .resources import GatewayType, LogLevel, PublicKeyFormat, PublicKeyCredential, DeviceCredential, FieldMask
from .utils import get_value
import os
from proto.datetime_helpers import DatetimeWithNanoseconds
import base64

def convertCredentialsFormatsFromString(credentials):
    if credentials is not None:
        # Converts public Key Format from string to object of class PublicKeyFormat
        for index, credential in enumerate(credentials):
            if 'publicKey' in credential:
                credential['publicKey']['format'] = PublicKeyFormat(credential['publicKey']['format'])
                credentials[index] = DeviceCredential(credential['publicKey'], credential['expirationTime'])
    return credentials

class Device():
    """
    Data class for Clearblade Device
    """
    # TODO: find a better way to construct the Device object. I dont like so much parameter in a constructor
    # From google SDK docs: The field ``name`` must be empty. The server generates ``name`` from the device 
    # registry ``id`` and the ``parent`` field.

    def __init__(self, id: str, num_id: str = None,
                 credentials: list = [], last_heartbeat_time: str = None, last_event_time: str = None,
                 last_state_time: str = None, last_config_ack_time: str = None,
                 last_config_send_time: str = None, blocked: bool = False,
                 last_error_time: str = None, last_error_status_code: dict = None,
                 config: dict = {"cloudUpdateTime":None, "version":""} ,
                 state: dict = {"updateTime":None, "binaryData":None},
                 log_level: str = LogLevel.LOG_LEVEL_UNSPECIFIED,
                 meta_data: dict = {},
                 gateway_config: dict = {"gatewayType": GatewayType.NON_GATEWAY}) -> None:

        self._id = id
        self._name = ''
        self._num_id = num_id
        self._credentials = credentials
        self._last_heartbeat_time = last_heartbeat_time
        self._last_event_time = last_event_time
        self._last_state_time = last_state_time
        self._last_config_ack_time = last_config_ack_time
        self._last_config_send_time = last_config_send_time
        self._blocked = blocked
        self._last_error_time = last_error_time
        self._last_error_status_code = last_error_status_code
        self._config = config
        self._state = state
        self._log_level = log_level
        self._meta_data = meta_data
        self._gateway_config = gateway_config

    @staticmethod
    def from_json(json):
        lastHeartbeatTimeFromJson = get_value(json,'lastHeartbeatTime')
        lastEventTimeFromJson = get_value(json, 'lastEventTime')
        lastStateTimeFromJson = get_value(json, 'lastStateTime')
        lastConfigAckTimeFromJson = get_value(json, 'lastConfigAckTime')
        lastConfigSendTimeFromJson = get_value(json, 'lastConfigSendTime')
        lastErrorTimeFromJson = get_value(json, 'lastErrorTime')
        configFromJson = get_value(json, 'config')
        stateFromJson = get_value(json, 'state')
        
        convert_times_to_datetime_with_nanoseconds = (False if os.environ.get("BINARYDATA_AND_TIME_GOOGLE_FORMAT") == None else os.environ.get("BINARYDATA_AND_TIME_GOOGLE_FORMAT").lower() == "true")
        if convert_times_to_datetime_with_nanoseconds:
            last_heartbeat_time = None if lastHeartbeatTimeFromJson in [None, ""] else DatetimeWithNanoseconds.from_rfc3339(lastHeartbeatTimeFromJson)
            last_event_time = None if lastEventTimeFromJson in [None, ""] else DatetimeWithNanoseconds.from_rfc3339(lastEventTimeFromJson)
            last_state_time = None if lastStateTimeFromJson in [None, ""] else DatetimeWithNanoseconds.from_rfc3339(lastStateTimeFromJson)
            last_config_ack_time = None if lastConfigAckTimeFromJson in [None, ""] else DatetimeWithNanoseconds.from_rfc3339(lastConfigAckTimeFromJson)
            last_config_send_time = None if lastConfigSendTimeFromJson in [None, ""] else DatetimeWithNanoseconds.from_rfc3339(lastConfigSendTimeFromJson)
            last_error_time = None if lastErrorTimeFromJson in [None, ""] else DatetimeWithNanoseconds.from_rfc3339(lastErrorTimeFromJson) 
        else:
            last_heartbeat_time = lastHeartbeatTimeFromJson
            last_event_time = lastEventTimeFromJson
            last_state_time = lastStateTimeFromJson
            last_config_ack_time = lastConfigAckTimeFromJson
            last_config_send_time = lastConfigSendTimeFromJson
            last_error_time = lastErrorTimeFromJson

        theDevice = Device(
            id=get_value(json, 'id'),
            num_id=get_value(json, 'numId'),
            credentials=convertCredentialsFormatsFromString(get_value(json, 'credentials')),
            last_heartbeat_time=last_heartbeat_time,
            last_event_time=last_event_time,
            last_state_time=last_state_time,
            last_config_ack_time=last_config_ack_time,
            last_config_send_time=last_config_send_time,
            last_error_time=last_error_time,
            blocked=get_value(json, 'blocked'),
            last_error_status_code=get_value(json, 'lastErrorStatus'),
            config=DeviceConfig.from_json(configFromJson),
            state=DeviceState.from_json(stateFromJson),
            log_level=get_value(json, 'logLevel'),
            meta_data=get_value(json, 'metadata'),
            gateway_config=get_value(json, 'gatewayConfig')
        )

        #Since _name is a private attribute, we have to populate it like this
        #because we don't allow "name" to be passed in the constructor

        theDevice._name=get_value(json, 'name')
        
        return theDevice

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def num_id(self):
        return self._num_id

    @property
    def credentials(self):
        return self._credentials

    @credentials.setter
    def credentials(self, credentials):
        self._credentials = credentials

    @property
    def last_error_status(self):
        return self._last_error_status_code

    @property
    def config(self):
        return self._config

    @property
    def state(self):
        return self._state

    @property
    def log_level(self):
        return self._log_level
    
    @log_level.setter
    def log_level(self, log_level):
        self._log_level = log_level

    @property
    def meta_data(self):
        return self._meta_data

    @meta_data.setter
    def meta_data(self, meta_data):
        self._meta_data = meta_data

    @property
    def gateway_config(self):
        return self._gateway_config

    @gateway_config.setter
    def gateway_config(self, gateway_config):
        self._gateway_config = gateway_config

    @property
    def log_level(self):
        return self._log_level

    @property
    def last_heartbeat_time(self):
        return self._last_heartbeat_time
    
    @property
    def last_event_time(self):
        return self._last_event_time
    
    @property
    def last_state_time(self):
        return self._last_state_time

    @property
    def last_config_ack_time(self):
        return self._last_config_ack_time

    @property
    def last_config_send_time(self):
        return self._last_config_send_time
    
    @property
    def last_error_time(self):
        return self._last_error_time

    @property
    def blocked(self):
        return self._blocked

    @blocked.setter
    def blocked(self, blocked):
        self._blocked = blocked


# classes to mock googles request & response

class DeviceState():
    def __init__(self, update_time: str = None, binary_data:str = None) -> None:
        self.updateTime = update_time
        self.binaryData = binary_data

    def __getitem__(self, arg):
        return getattr(self, arg)

    def get(self, arg):
        return getattr(self, arg)

    @property
    def update_time(self):
        return self.updateTime

    @property
    def binary_data(self):
        return self.binaryData

    @staticmethod
    def from_json(response_json):
        updateTimeFromJson = get_value(response_json, 'updateTime')
        binaryDataFromJson = get_value(response_json, 'binaryData')
        
        convert_times_to_datetime_with_nanoseconds = (False if os.environ.get("BINARYDATA_AND_TIME_GOOGLE_FORMAT") == None else os.environ.get("BINARYDATA_AND_TIME_GOOGLE_FORMAT").lower() == "true")
        if convert_times_to_datetime_with_nanoseconds:
            update_time = None if updateTimeFromJson in [None, ""] else DatetimeWithNanoseconds.from_rfc3339(updateTimeFromJson)
        else:
            update_time = updateTimeFromJson

        if (binaryDataFromJson not in [None, ""]):
            convert_binarydata_to_bytes = (False if os.environ.get("BINARYDATA_AND_TIME_GOOGLE_FORMAT") == None else os.environ.get("BINARYDATA_AND_TIME_GOOGLE_FORMAT").lower() == "true")
            if convert_binarydata_to_bytes:
                binary_data = binaryDataFromJson.encode('utf-8')
            else:
                binary_data = binaryDataFromJson
        else:
            binary_data = binaryDataFromJson

        return DeviceState(update_time=update_time,
                           binary_data=binary_data)
       
class Request():
    def __init__(self, parent) -> None:
        self._parent = parent

    @property
    def parent(self):
        return self._parent


class SendCommandToDeviceRequest(Request):
    def __init__(self, name: str = None,
                 binary_data: bytes = None,
                 subfolder: str = None) -> None:
        super().__init__(name)
        self._binary_data = binary_data
        self._subfolder = subfolder

    @property
    def binary_data(self):
        return self._binary_data

    @property
    def sub_folder(self):
        return self._subfolder


class CreateDeviceRequest(Request):
    def __init__(self, parent: str = None,
                 device: Device = None) -> None:
        super().__init__(parent)
        self._device = device

    @property
    def device(self):
        return self._device


class ModifyCloudToDeviceConfigRequest(Request):
    def __init__(self, name:str = None ,
                 version_to_update: str = "",
                 binary_data: bytes = None) -> None:
        super().__init__(name)
        self._version_to_update = version_to_update
        self._binary_data = binary_data

    @property
    def version_to_update(self):
        return self._version_to_update

    @property
    def binary_data(self):
        return self._binary_data


class DeviceConfig(Request):
    def __init__(self, name,
                 version,
                 cloud_update_time,
                 device_ack_time,
                 binary_data) -> None:
        super().__init__(name)
        self._version = version
        self.cloudUpdateTime = cloud_update_time
        self.deviceAckTime = device_ack_time
        self.binaryData = binary_data

    def __getitem__(self, arg):
        return getattr(self, arg)

    def get(self, arg):
        return getattr(self, arg)

    @property
    def version(self):
        return self._version

    @property
    def cloud_update_time(self):
        return self.cloudUpdateTime

    @property
    def device_ack_time(self):
        return self.deviceAckTime

    @property
    def binary_data(self):
        return self.binaryData

    @staticmethod
    def from_json(json):
        cloudUpdateTimeFromJson = get_value(json,'cloudUpdateTime')
        deviceAckTimeFromJson = get_value(json, 'deviceAckTime')
        binaryDataFromJson = get_value(json,'binaryData')

        convert_times_to_datetime_with_nanoseconds = (False if os.environ.get("BINARYDATA_AND_TIME_GOOGLE_FORMAT") == None else os.environ.get("BINARYDATA_AND_TIME_GOOGLE_FORMAT").lower() == "true")
        if convert_times_to_datetime_with_nanoseconds:
            cloud_update_time = None if cloudUpdateTimeFromJson in [None, ""] else DatetimeWithNanoseconds.from_rfc3339(cloudUpdateTimeFromJson)
            device_ack_time = None if deviceAckTimeFromJson in [None, ""] else DatetimeWithNanoseconds.from_rfc3339(deviceAckTimeFromJson)
        else:
            cloud_update_time = cloudUpdateTimeFromJson
            device_ack_time = deviceAckTimeFromJson
        
        if binaryDataFromJson not in [None, ""]:
            convert_binarydata_to_bytes = (False if os.environ.get("BINARYDATA_AND_TIME_GOOGLE_FORMAT") == None else os.environ.get("BINARYDATA_AND_TIME_GOOGLE_FORMAT").lower() == "true")
            if convert_binarydata_to_bytes:
                binary_data = base64.b64decode(binaryDataFromJson.encode('utf-8'))
            else:
                binary_data = binaryDataFromJson
        else:
            binary_data = binaryDataFromJson

        return DeviceConfig(name='',
                            version=get_value(json, 'version'),
                            cloud_update_time=cloud_update_time,
                            device_ack_time=device_ack_time,
                            binary_data=binary_data)


class DeleteDeviceRequest(Request):
    def __init__(self, name: str = None) -> None:
        super().__init__(name)


class GetDeviceRequest(Request):
    def __init__(self, name: str = None) -> None:
        super().__init__(name)


class BindUnBindGatewayDeviceRequest(Request):
    def __init__(self, parent:str = None,
                 deviceId: str = None,
                 gatewayId: str = None) -> None:
        self._parent = parent
        self._deviceid = deviceId
        self._gatewayid = gatewayId

    @property
    def parent(self):
        return self._parent

    @property
    def deviceId(self):
        return self._deviceid

    @property
    def gatewayId(self):
        return self._gatewayid


class BindDeviceToGatewayRequest(BindUnBindGatewayDeviceRequest):
    def __init__(self, parent: str = None,
                 deviceId: str = None,
                 gatewayId: str = None) -> None:
        super().__init__(parent, deviceId, gatewayId)


class UnbindDeviceFromGatewayRequest(BindUnBindGatewayDeviceRequest):
    def __init__(self, parent: str = None,
                 deviceId: str = None,
                 gatewayId: str = None) -> None:
        super().__init__(parent, deviceId, gatewayId)


class ListDeviceStatesRequest(Request):
    def __init__(self, name: str = None,
                 numStates: int = None) -> None:
        super().__init__(name)
        self._numStates = numStates

    @property
    def numStates(self):
        return self._numStates


class ListDeviceStatesResponse():
    def __init__(self, device_states:List[DeviceState] = []) -> None:
        self._device_states = device_states

    @property
    def device_states(self):
        return self._device_states

    @staticmethod
    def from_json(response_json):
        device_states_json = get_value(response_json, 'deviceStates')
        device_states = []
        for device_state_json in device_states_json:
            device_state = DeviceState.from_json(device_state_json)
            device_states.append(device_state)

        return ListDeviceStatesResponse(device_states=device_states)


class ListDeviceConfigVersionsRequest(Request):
    def __init__(self, name: str = None,
                 numVersions: int = None) -> None:
        super().__init__(name)
        self._numversions = numVersions

    @property
    def numVersions(self):
        return self._numversions


class ListDeviceConfigVersionsResponse():
    def __init__(self, device_configs) -> None:
        self._device_configs = device_configs

    @property
    def device_configs(self):
        return self._device_configs

    @staticmethod
    def from_json(response_json):
        device_configs_json = response_json['deviceConfigs']
        deviceConfigs = []
        for device_config_json in device_configs_json:
            deviceConfig = DeviceConfig.from_json(device_config_json)
            deviceConfigs.append(deviceConfig)

        return ListDeviceConfigVersionsResponse(device_configs=deviceConfigs)


class UpdateDeviceRequest(Request):
    def __init__(self, parent: str = None, device: Device = None, updateMask: str = None) -> None:
        self._parent = parent
        self._device = device
        self._update_mask = updateMask

    def _prepare_params_body_for_update(self):
        params = {'name': self._device.id}
        if self._update_mask is not None:
            params['updateMask'] = self._update_mask

        body = {}
        if self._device.log_level is not None:
            body['logLevel'] = self._device.log_level
        if self._device.gateway_config is not None:
            body['gatewayConfig'] = self._device.gateway_config
        if self._device.meta_data is not None:
            body['metadata'] = self._device.meta_data
        if self._device._blocked is not None:
            body['blocked'] = self._device._blocked
        if self._device._credentials is not None:
            body['credentials'] = DeviceCredential.convert_credentials_for_create_update(self._device._credentials)

        return params, body


class ListDevicesRequest(Request):
    def __init__(self, parent:str = None ,
                 deviceNumIds: str = None,
                 deviceIds: str = None,
                 fieldMask: str = None,
                 gatewayListOptions: dict = None,
                 pageSize: int = None,
                 pageToken: str = None) -> None:
        self._parent = parent
        self._device_num_ids = deviceNumIds
        self._device_ids = deviceIds
        self._field_mask = fieldMask
        self._gateway_list_options = gatewayListOptions
        self._page_size = pageSize
        self._page_token = pageToken

    @property
    def parent(self):
        return self._parent

    @property
    def device_num_ids(self):
        return self._device_num_ids

    @property
    def device_ids(self):
        return self._device_ids

    @property
    def field_mask(self):
        return self._field_mask

    @property
    def gateway_list_options(self):
        return self._gateway_list_options

    @property
    def page_size(self):
        return self._page_size

    @property
    def page_token(self):
        return self._page_token

    @page_token.setter
    def page_token(self, page_token):
        self._page_token = page_token

    def _prepare_params_for_list(self):
        params = {'parent':self.parent}
        if self.page_size:
            params['pageSize'] = self.page_size
        if self.device_num_ids:
            params['deviceNumIds'] = self.device_num_ids
        if self.device_ids:
            params['deviceIds'] = self.device_ids
        if self.field_mask:
            params['fieldMask'] = FieldMask.convert_fieldmask_for_list(self.field_mask)
        if self.gateway_list_options :
            if 'associationsDeviceId' in self.gateway_list_options:
                params['gatewayListOptions.associationsDeviceId'] = self.gateway_list_options['associationsDeviceId']
            if 'associationsGatewayId' in self.gateway_list_options:
                params['gatewayListOptions.associationsGatewayId'] = self.gateway_list_options['associationsGatewayId']
            if 'gatewayType' in self.gateway_list_options:
                params['gatewayListOptions.gatewayType'] = self.gateway_list_options['gatewayType']
        if self.page_token:
            params['pageToken'] = self.page_token

        return params


class ListDevicesResponse():

    def __init__(self, devices, next_page_token) -> None:
        self._devices = devices
        self._next_page_token = next_page_token

    @property
    def raw_pages(self):
        return self

    @property
    def devices(self):
        return self._devices

    @property
    def next_page_token(self):
        return self._next_page_token

    @staticmethod
    def from_json(devices_list_json):
        devices = []
        next_page_token = None
        if 'devices' in devices_list_json:
            devices_json = devices_list_json['devices']
            for device_json in devices_json:
                device = Device.from_json(device_json)
                devices.append(device)

        if 'nextPageToken' in devices_list_json:
            next_page_token = devices_list_json['nextPageToken']

        return ListDevicesResponse(devices=devices, next_page_token=next_page_token)


"""
"Copyright 2023 ClearBlade Inc."

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2023 ClearBlade Inc.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .config_manager import ClearBladeConfigManager
from .http_client import AsyncClient, SyncClient
from .pagers import ListDeviceRegistriesAsyncPager, ListDeviceRegistryPager
from .registry_types import *


class ClearBladeRegistryManager():
    def __init__(self) -> None:
        self._cb_config = ClearBladeConfigManager()

    def _create_registry_body(self, registry: DeviceRegistry) :
        registry_json = {'id':registry.id}
        if registry.credentials:
            registry_json['credentials']=registry.credentials
        if registry.http_config:
            registry_json['httpConfig']=registry.http_config
        if registry.mqtt_config:
            registry_json['mqttConfig']=registry.mqtt_config
        if registry.state_notification_config:
            registry_json['stateNotificationConfig']=registry.state_notification_config
        if registry.event_notification_configs:
            registry_json['eventNotificationConfigs']=registry.event_notification_configs
        if registry.log_level:
            registry_json['logLevel']=registry.log_level
        return registry_json

    def _prepare_params_for_registry_list(self, request:ListDeviceRegistriesRequest):
        request_params = {'parent':request.parent}
        if request.page_size:
            request_params['page_size'] = request.page_size
        if request.page_token:
            request_params['page_token'] = request.page_token
        return request_params

    def create(self,
        request: CreateDeviceRegistryRequest = None)->DeviceRegistry:
        body = self._create_registry_body(request.device_registry)
        params = {'parent':request.parent}
        sync_client = SyncClient(clearblade_config=self._cb_config.admin_config)
        response = sync_client.post(api_name = "cloudiot",request_body=body,request_params=params)
        return DeviceRegistry.from_json(response.json())

    async def create_async(self,
        request: CreateDeviceRegistryRequest = None)->DeviceRegistry:
        body = self._create_registry_body(request.device_registry)
        params = {'parent':request.parent}
        async_client = AsyncClient(clearblade_config=self._cb_config.admin_config)
        response = await async_client.post(api_name = "cloudiot",request_body=body,request_params=params)
        return DeviceRegistry.from_json(response.json())

    def get(self,
            request: GetDeviceRegistryRequest = None):
        if request.name == None:
            raise Exception('Registry Path is not present in the request')

        clearblade_config = self._cb_config.regional_config(request.name)

        sync_client = SyncClient(clearblade_config=clearblade_config)
        response = sync_client.get(api_name = "cloudiot")
        return DeviceRegistry.from_json(response.json())

    async def get_async(self,
                        request: GetDeviceRegistryRequest = None):
        
        if request.name == None:
            raise Exception('Registry Path is not present in the request')

        clearblade_config=await self._cb_config.regional_config_async(request.name)
        async_client = AsyncClient(clearblade_config=clearblade_config)
        response = await async_client.get(api_name = "cloudiot")
        return DeviceRegistry.from_json(response.json())

    def get_device_registry_list(self, request: ListDeviceRegistriesRequest):
        request_params = self._prepare_params_for_registry_list(request=request)
        sync_client = SyncClient(clearblade_config=self._cb_config.admin_config)
        response = sync_client.get(api_name = "cloudiot",request_params=request_params)

        if response.status_code == 200:
            return ListDeviceRegistriesResponse.from_json(response.json())
        return None

    def list(self,
            request: ListDeviceRegistriesRequest = None):

        list_device_registry_response = self.get_device_registry_list(request=request)
        if list_device_registry_response:
            return ListDeviceRegistryPager(method=self.get_device_registry_list,
                                           request=request,
                                           response=list_device_registry_response)

    async def get_device_registry_list_async(self, request: ListDeviceRegistriesRequest):
        request_params = self._prepare_params_for_registry_list(request=request)
        async_client = AsyncClient(clearblade_config=self._cb_config.admin_config)
        response = await async_client.get(api_name = "cloudiot",request_params=request_params)

        if response.status_code == 200:
            return ListDeviceRegistriesResponse.from_json(response.json())
        return None

    async def list_async(self, request: ListDeviceRegistriesRequest):
        list_device_registry_response = await self.get_device_registry_list_async(request=request)
        if list_device_registry_response:
            return ListDeviceRegistriesAsyncPager(method=self.get_device_registry_list_async,
                                                  request=request,
                                                  response=list_device_registry_response)
        return None

    def delete(self,
            request: DeleteDeviceRegistryRequest = None):
        params = {'name':request.name}
        sync_client = SyncClient(clearblade_config=self._cb_config.admin_config)
        response = sync_client.delete(api_name = "cloudiot",request_params=params)
        return response

    async def delete_async(self,
            request: DeleteDeviceRegistryRequest = None):
        params = {'name':request.name}
        async_client = AsyncClient(clearblade_config=self._cb_config.admin_config)
        response = await async_client.delete(api_name = "cloudiot",request_params=params)
        return response

    def patch(self,
        request: UpdateDeviceRegistryRequest = None)->DeviceRegistry:
        if request.name == None:
            raise Exception('Registry Path is not present in the request')

        clearblade_config = self._cb_config.regional_config(request.name)
        
        body = self._create_registry_body(request.device_registry)
        params = {'name':request.name, 'updateMask': request.update_mask}
        sync_client = SyncClient(clearblade_config=clearblade_config)
        response = sync_client.patch(api_name = "cloudiot",request_body=body,request_params=params)
        return DeviceRegistry.from_json(response.json())

    async def patch_async(self,
        request: UpdateDeviceRegistryRequest = None)->DeviceRegistry:
        
        if request.name == None:
            raise Exception('Registry Path is not present in the request')

        body = self._create_registry_body(request.device_registry)
        params = {'name':request.name, 'updateMask': request.update_mask}
        clearblade_config=await self._cb_config.regional_config_async(request.name)
        async_client = AsyncClient(clearblade_config=clearblade_config)
        response = await async_client.patch(api_name = "cloudiot",request_body=body,request_params=params)
        return DeviceRegistry.from_json(response.json())


"""
"Copyright 2023 ClearBlade Inc."

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2023 ClearBlade Inc.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from enum import Enum
from datetime import datetime

class MqttState():
    r"""Indicates whether an MQTT connection is enabled or disabled.
    See the field description for details.
    """
    MQTT_STATE_UNSPECIFIED = "MQTT_STATE_UNSPECIFIED"
    MQTT_ENABLED = "MQTT_ENABLED"
    MQTT_DISABLED = "MQTT_DISABLED"


class HttpState():
    r"""Indicates whether DeviceService (HTTP) is enabled or disabled
    for the registry. See the field description for details.
    """
    HTTP_STATE_UNSPECIFIED = "HTTP_STATE_UNSPECIFIED"
    HTTP_ENABLED = "HTTP_ENABLED"
    HTTP_DISABLED = "HTTP_DISABLED"


class LogLevel:
    r"""**Beta Feature**

    The logging verbosity for device activity. Specifies which events
    should be written to logs. For example, if the LogLevel is ERROR,
    only events that terminate in errors will be logged. LogLevel is
    inclusive; enabling INFO logging will also enable ERROR logging.
    """
    LOG_LEVEL_UNSPECIFIED = "LOG_LEVEL_UNSPECIFIED"
    NONE = "NONE"
    ERROR = "ERROR"
    INFO = "INFO"
    DEBUG = "DEBUG"


class GatewayType:
    r"""Gateway type."""
    GATEWAY_TYPE_UNSPECIFIED = "GATEWAY_TYPE_UNSPECIFIED"
    GATEWAY = "GATEWAY"
    NON_GATEWAY = "NON_GATEWAY"


class GatewayAuthMethod():
    r"""The gateway authorization/authentication method. This setting
    determines how Cloud IoT Core authorizes/authenticate devices to
    access the gateway.
    """
    GATEWAY_AUTH_METHOD_UNSPECIFIED = "GATEWAY_AUTH_METHOD_UNSPECIFIED"
    ASSOCIATION_ONLY = "ASSOCIATION_ONLY"
    DEVICE_AUTH_TOKEN_ONLY = "DEVICE_AUTH_TOKEN_ONLY"
    ASSOCIATION_AND_DEVICE_AUTH_TOKEN = "ASSOCIATION_AND_DEVICE_AUTH_TOKEN"


class PublicKeyCertificateFormat():
    r"""The supported formats for the public key."""
    UNSPECIFIED_PUBLIC_KEY_CERTIFICATE_FORMAT = "UNSPECIFIED_PUBLIC_KEY_CERTIFICATE_FORMAT"
    X509_CERTIFICATE_PEM = "X509_CERTIFICATE_PEM"


class PublicKeyFormat(Enum):
    r"""The supported formats for the public key."""
    UNSPECIFIED_PUBLIC_KEY_FORMAT = "UNSPECIFIED_PUBLIC_KEY_FORMAT"
    RSA_PEM = "RSA_PEM"
    RSA_X509_PEM = "RSA_X509_PEM"
    ES256_PEM = "ES256_PEM"
    ES256_X509_PEM = "ES256_X509_PEM"

class PublicKeyCredential():
    def __init__(self, format: PublicKeyFormat, key: bytes):
        self.format = format
        self.key = key
    
    def __getitem__(self, arg):
        return getattr(self, arg)

    def get(self, arg):
        return getattr(self, arg)


class DeviceCredential():
    def __init__(self, public_key, expiration_time=''):
        if isinstance(public_key, dict):
            self.publicKey = PublicKeyCredential(public_key['format'], public_key['key'])
        else:
            self.publicKey = public_key
        self.expirationTime = expiration_time

    def __getitem__(self, arg):
        return getattr(self, arg)

    def get(self, arg):
        return getattr(self, arg)

    @property
    def public_key(self):
        return self.publicKey

    @property
    def expiration_time(self):
        return self.expirationTime

    @classmethod
    def convert_credentials_for_create_update(cls, credentials):
        for index, credential in enumerate(credentials):
            # Convert credential to dict if it is not
            updateDeviceCredential = False
            if (isinstance(credential, DeviceCredential)):
                credential = credential.__dict__
                updateDeviceCredential = True
            
            if 'publicKey' in credential:
                if (isinstance(credential['publicKey'], PublicKeyCredential)):
                    credential['publicKey'] = credential['publicKey'].__dict__
                # Convert PublicKeyFormat to string
                credential['publicKey']['format'] = PublicKeyFormat(credential['publicKey']['format']).value
                updateDeviceCredential = True
            
            if 'expirationTime' in credential:
                if (isinstance(credential['expirationTime'], datetime)):
                    credential['expirationTime'] = credential['expirationTime'].strftime('%Y-%m-%dT%H:%M:%SZ')
                    updateDeviceCredential = True
    
            if updateDeviceCredential:
                credentials[index] = credential
        
        return credentials

class FieldMask():
    def __init__(self, paths: [str]):
        self.paths = paths
    
    def __getitem__(self, arg):
        return getattr(self, arg)

    def get(self, arg):
        return getattr(self, arg)

    @classmethod
    def convert_fieldmask_for_list(cls, field_mask):
        if (isinstance(field_mask, FieldMask)):
            field_mask = field_mask.__dict__
            if 'paths' in field_mask:
                field_mask = field_mask['paths']
        return field_mask



"""
"Copyright 2023 ClearBlade Inc."

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2023 ClearBlade Inc.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Any

def find_project_region_registry_from_parent(parent):
    #projects/ingressdevelopmentenv/locations/us-central1/registries/gargi_python
    if not parent:
        return None

    list = parent.split("/")
    project_region_registry_dict = {}    
    if "projects" in list and len(list) >= 2:
        project_region_registry_dict['project'] = list[1]
    if "locations" in list and len(list) >= 4:
        project_region_registry_dict['location'] = list[3]
    if "registries" in list and len(list) >= 6:
        project_region_registry_dict['registry'] = list[5]

    return project_region_registry_dict

def get_value(json_data, key):
    if json_data is not None:
        if key in json_data:
            return json_data[key]
    return None

class SingletonMetaClass(type):

    _instances = {}

    def __call__(cls, *args: Any, **kwds: Any) -> Any:
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwds)
            cls._instances[cls] = instance
        return cls._instances[cls]





# -*- coding: utf-8 -*-
#
# Copyright 2023 ClearBlade Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time

from google.cloud import iot_v1


class TestSystemDeviceManager(object):
    def test_list_device_registries(self):
        project_id = os.environ["PROJECT_ID"]

        client = iot_v1.DeviceManagerClient()
        parent = f"projects/{project_id}/locations/us-central1"
        response = client.list_device_registries(request={"parent": parent})



"""
"Copyright 2023 ClearBlade Inc."

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2023 ClearBlade Inc.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Copyright 2018 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .device_types import Device, ListDevicesResponse, ListDevicesRequest
from .registry_types import DeviceRegistry, ListDeviceRegistriesRequest, ListDeviceRegistriesResponse
from typing import Any, Awaitable, AsyncIterator, Callable

class ListDeviceRegistryPager:
    def __init__(self, method, 
                 request: ListDeviceRegistriesRequest, 
                 response:ListDeviceRegistriesResponse) -> None:
        self._method = method
        self._request = request
        self._response = response

    def __getattr__(self, name: str) -> Any:
        return getattr(self._response, name)

    @property
    def pages(self):
        yield self._response
        is_next_page_token = self._response.next_page_token and self._response.next_page_token != 0
        while is_next_page_token:
            # call the method to fetch the next page.
            self._request.page_token = self._response.next_page_token
            self._response = self._method(request = self._request)
            yield self._response

    def __iter__(self):
        for page in self.pages:
            yield from page.device_registries

    def __repr__(self) -> str:
        return "{0}<{1!r}>".format(self.__class__.__name__, self._response)

class ListDeviceRegistriesAsyncPager:

    def __init__(
        self,
        method: Callable[..., Awaitable[ListDeviceRegistriesResponse]],
        request: ListDeviceRegistriesRequest,
        response: ListDeviceRegistriesResponse
        ):
        self._method = method
        self._request = request
        self._response = response

    def __getattr__(self, name: str) -> Any:
        return getattr(self._response, name)

    @property
    async def pages(self) -> AsyncIterator[ListDeviceRegistriesResponse]:
        yield self._response
        is_next_page_token = self._response.next_page_token and self._response.next_page_token != 0

        while is_next_page_token:
            self._request.page_token = self._response.next_page_token
            self._response = await self._method(self._request)
            yield self._response

    def __aiter__(self) -> AsyncIterator[DeviceRegistry]:
        async def async_generator():
            async for page in self.pages:
                for response in page.device_registries:
                    yield response

        return async_generator()

    def __repr__(self) -> str:
        return "{0}<{1!r}>".format(self.__class__.__name__, self._response)

class ListDevicesPager:

    def __init__(self, method: Callable[..., ListDevicesResponse],
                 request: ListDevicesRequest,
                 response:ListDevicesResponse) -> None:
        self._method = method
        self._request = request
        self._response = response

    @property
    def pages(self):
        yield self._response
        while self._response.next_page_token:
            # call the method to fetch the next page.
            self._request.page_token = self._response.next_page_token
            self._response = self._method(request = self._request)
            yield self._response

    def __iter__(self):
        for page in self.pages:
            yield from page.devices

    def __repr__(self) -> str:
        return "{0}<{1!r}>".format(self.__class__.__name__, self._response)

class ListDevicesAsyncPager:
    """A pager for iterating through ``list_devices`` requests.

    This class thinly wraps an initial
    :class:`google.cloud.iot_v1.types.ListDevicesResponse` object, and
    provides an ``__aiter__`` method to iterate through its
    ``devices`` field.

    If there are more pages, the ``__aiter__`` method will make additional
    ``ListDevices`` requests and continue to iterate
    through the ``devices`` field on the
    corresponding responses.

    All the usual :class:`google.cloud.iot_v1.types.ListDevicesResponse`
    attributes are available on the pager. If multiple requests are made, only
    the most recent response is retained, and thus used for attribute lookup.
    """

    def __init__(
        self,
        method: Callable[..., Awaitable[ListDevicesResponse]],
        request: ListDevicesRequest,
        response: ListDevicesResponse,
        ):
        """Instantiates the pager.

        Args:
            method (Callable): The method that was originally called, and
                which instantiated this pager.
            request (ListDevicesRequest):
                The initial request object.
            response (ListDevicesResponse):
                The initial response object.
        """
        self._method = method
        self._request = request
        self._response = response

    def __getattr__(self, name: str) -> Any:
        return getattr(self._response, name)

    @property
    async def pages(self) -> AsyncIterator[ListDevicesResponse]:
        yield self._response
        while self._response.next_page_token:
            self._request.page_token = self._response.next_page_token
            self._response = await self._method(self._request)
            yield self._response

    def __aiter__(self) -> AsyncIterator[Device]:
        async def async_generator():
            async for page in self.pages:
                for response in page.devices:
                    yield response

        return async_generator()

    def __repr__(self) -> str:
        return "{0}<{1!r}>".format(self.__class__.__name__, self._response)